package ity.opencraft.data;

import ity.opencraft.backends.Backend;
import ity.opencraft.backends.BackendImpl;
import org.joml.Vector2f;
import org.joml.Vector3f;

import java.util.List;

import static org.lwjgl.opengl.GL30.*;

@BackendImpl(Backend.GL)
public class BlockMeshGL extends BlockMesh {
    private int vaoID;

    private int pboID;
    private int nboID;
    private int tboID;

    private int lboID;

    private int iboID;

    private int length;

    @Override
    public void init() {
        vaoID = glGenVertexArrays();
        glBindVertexArray(vaoID);

        pboID = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, pboID);
        glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        nboID = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, nboID);
        glVertexAttribPointer(1, 3, GL_FLOAT, false, 0, 0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        tboID = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, tboID);
        glVertexAttribPointer(2, 2, GL_FLOAT, false, 0, 0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        lboID = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, lboID);
        glVertexAttribPointer(3, 1, GL_FLOAT, false, 0, 0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        iboID = glGenBuffers();

        glBindVertexArray(0);
    }

    @Override
    public void upload(float[] pos, float[] norm, float[] texIds, float[] light, int[] inds) {
        this.length = inds.length;

        glBindVertexArray(vaoID);

        glBindBuffer(GL_ARRAY_BUFFER, pboID);
        glBufferData(GL_ARRAY_BUFFER, pos, GL_STATIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        glBindBuffer(GL_ARRAY_BUFFER, nboID);
        glBufferData(GL_ARRAY_BUFFER, norm, GL_STATIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        glBindBuffer(GL_ARRAY_BUFFER, tboID);
        glBufferData(GL_ARRAY_BUFFER, texIds, GL_STATIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        glBindBuffer(GL_ARRAY_BUFFER, lboID);
        glBufferData(GL_ARRAY_BUFFER, light, GL_STATIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboID);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, inds, GL_STATIC_DRAW);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

        glBindVertexArray(0);
    }

    @Override
    public void upload(List<Vector3f> pos, List<Vector3f> norm, List<Vector2f> texIds, List<Float> light, List<Integer> inds) {
        float[] pBuf = new float[pos.size() * 3];
        for(int i = 0; i < pos.size(); i++) {
            pBuf[i*3] = pos.get(i).x;
            pBuf[i*3+1] = pos.get(i).y;
            pBuf[i*3+2] = pos.get(i).z;
        }

        float[] nBuf = new float[norm.size() * 3];
        for(int i = 0; i < norm.size(); i++) {
            nBuf[i*3] = norm.get(i).x;
            nBuf[i*3+1] = norm.get(i).y;
            nBuf[i*3+2] = norm.get(i).z;
        }

        float[] tBuf = new float[texIds.size() * 2];
        for(int i = 0; i < texIds.size(); i++) {
            tBuf[i*2] = texIds.get(i).x;
            tBuf[i*2+1] = texIds.get(i).y;
        }

        float[] lBuf = new float[light.size()];
        for(int i = 0; i < light.size(); i++) {
            lBuf[i] = light.get(i);
        }

        int[] iBuf = new int[inds.size()];
        for(int i = 0; i < inds.size(); i++) {
            iBuf[i] = inds.get(i);
        }

        upload(pBuf, nBuf, tBuf, lBuf, iBuf);
    }

    @Override
    public void bind() {
        glBindVertexArray(vaoID);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboID);
        glEnableVertexAttribArray(0);
        glEnableVertexAttribArray(1);
        glEnableVertexAttribArray(2);
        glEnableVertexAttribArray(3);
    }

    @Override
    public void render() {
        // TODO this is a bit hacky, idk...
        if(this.length > 0)
            glDrawElements(GL_TRIANGLES, this.length, GL_UNSIGNED_INT, 0);
    }

    @Override
    public void unbind() {
        glBindVertexArray(0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        glDisableVertexAttribArray(0);
        glDisableVertexAttribArray(1);
        glDisableVertexAttribArray(2);
        glDisableVertexAttribArray(3);
    }

    @Override
    public void destroy() {
        glDeleteVertexArrays(vaoID);
        glDeleteBuffers(pboID);
        glDeleteBuffers(nboID);
        glDeleteBuffers(tboID);
        glDeleteBuffers(lboID);
        glDeleteBuffers(iboID);
    }

    @Override
    public int getId() {
        return vaoID;
    }
}
