package ity.opencraft.data;

import ity.opencraft.backends.Backend;
import ity.opencraft.backends.BackendImpl;
import org.joml.Vector3f;
import static org.lwjgl.opengl.GL30.*;

import java.util.List;

@BackendImpl(Backend.GL)
public class IndexedMeshGL extends IndexedMesh {
    private int vaoID;

    private int pboID;

    private int iboID;

    private int length;

    public void init() {
        vaoID = glGenVertexArrays();
        glBindVertexArray(vaoID);

        pboID = glGenBuffers();
        glBindBuffer(GL_ARRAY_BUFFER, pboID);
        glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        iboID = glGenBuffers();

        glBindVertexArray(0);
    }

    public void upload(float[] pos, int[] inds) {
        this.length = inds.length;

        glBindVertexArray(vaoID);

        glBindBuffer(GL_ARRAY_BUFFER, pboID);
        glBufferData(GL_ARRAY_BUFFER, pos, GL_STATIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboID);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, inds, GL_STATIC_DRAW);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

        glBindVertexArray(0);
    }

    public void upload(List<Vector3f> pos, List<Integer> inds) {
        float[] pBuf = new float[pos.size() * 3];
        for(int i = 0; i < pos.size(); i++) {
            pBuf[i*3] = pos.get(i).x;
            pBuf[i*3+1] = pos.get(i).y;
            pBuf[i*3+2] = pos.get(i).z;
        }

        int[] iBuf = new int[inds.size()];
        for(int i = 0; i < inds.size(); i++) {
            iBuf[i] = inds.get(i);
        }

        upload(pBuf, iBuf);
    }

    public void bind() {
        glBindVertexArray(vaoID);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboID);
        glEnableVertexAttribArray(0);
    }

    public void render() {
        glDrawElements(GL_TRIANGLES, this.length, GL_UNSIGNED_INT, 0);
    }

    public void unbind() {
        glBindVertexArray(0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        glDisableVertexAttribArray(0);
    }

}
