package ity.opencraft.io;

import org.lwjgl.glfw.GLFW;
import org.lwjgl.glfw.GLFWWindowSizeCallback;

import static org.lwjgl.opengl.GL11.glViewport;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Window {
    private long wID;
    public boolean maximized;

    private Keyboard keyboard;
    private Mouse mouse;

    public int width, height;
    public void init(int sizeX, int sizeY, String title) {
        this.width = sizeX; // Ah java...
        this.height = sizeY;
        wID = GLFW.glfwCreateWindow(width, height, title, NULL, NULL);

        GLFW.glfwSetWindowSizeCallback(wID, new GLFWWindowSizeCallback() {
            @Override
            public void invoke(long window, int w, int h) {
                width = w;
                height = h;
                glViewport(0, 0, w, h);
            }
        });
        this.keyboard = new Keyboard();
        keyboard.init(wID);
        this.mouse = new Mouse();
        mouse.init(wID);
    }
    public void select() {
        GLFW.glfwMakeContextCurrent(wID);
        GLFW.glfwSwapInterval(1);
        new BackendCapabilities().initCaps();
    }
    public void swap() {
        glViewport(0, 0, width, height);
        GLFW.glfwSwapBuffers(wID);
    }

    public void fullscreen() {
        maximized = true;
        GLFW.glfwMaximizeWindow(wID);
    }

    public void restore() {
        maximized = false;
        GLFW.glfwRestoreWindow(wID);
    }

    public Keyboard getKeyboard() {
        return this.keyboard;
    }

    public Mouse getMouse() {
        return this.mouse;
    }

    public void setCursor(int cursor) {
        GLFW.glfwSetInputMode(wID, GLFW.GLFW_CURSOR, cursor);
    }

    /**
     * @return whether the close button or similar was pressed
     */ 
    public boolean shouldClose() {
        return GLFW.glfwWindowShouldClose(wID);
    }
}
