package ity.opencraft.mesh;

import ity.opencraft.world.Chunk;
import ity.opencraft.world.Block;
import ity.opencraft.world.Blocks;
import ity.opencraft.world.World;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;

import java.util.ArrayList;
import java.util.HashMap;

public class ChunkMesher {

    private static Block top(int x, int y, int z) {
        if(y == World.sizeY* Chunk.size) return Blocks.AIR;
        return World.getBlockAtLazy(x, y+1, z);
    }
    private static Block bottom(int x, int y, int z) {
        if(y == 0) return Blocks.AIR;
        return World.getBlockAtLazy(x, y-1, z);
    }
    private static Block left(int x, int y, int z) {
        if(x == 0) return Blocks.AIR;
        return World.getBlockAtLazy(x-1, y, z);
    }
    private static Block right(int x, int y, int z) {
        if(x == World.sizeXZ*Chunk.size) return Blocks.AIR;
        return World.getBlockAtLazy(x+1, y, z);
    }
    private static Block back(int x, int y, int z) {
        if(z == 0) return Blocks.AIR;
        return World.getBlockAtLazy(x, y, z-1);
    }
    private static Block front(int x, int y, int z) {
        if(z == World.sizeXZ*Chunk.size) return Blocks.AIR;
        return World.getBlockAtLazy(x, y, z+1);
    }


    /**
     * Fill the given parameters with the mesh of the chunk at the given position
     * @param posA The ArrayList to fill with positions
     * @param normA The ArrayList tof ill with normals
     * @param inxA The ArrayList to fill with indices
     * @param texA The ArrayList to fill with texture ids
     * @param loc The location of the chunk, in chunk coordinates.
     */
    public static void meshChunkWithSolid(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, ArrayList<Float> light, Vector3i loc) {
        for(int i = 0; i < Chunk.size; i++) {
            for(int j = 0; j < Chunk.size; j++) {
                for(int k = 0; k < Chunk.size; k++) {
                    int wPosX = i + loc.x*Chunk.size, wPosY = j + loc.y*Chunk.size, wPosZ = k + loc.z*Chunk.size;
                    Block block = World.getBlockAtLazy(wPosX, wPosY, wPosZ);
                    if(block.visible() && !block.isBlend()) {
                        switch(block.modelType()) {
                            case CUBE:
                                CubeMesher.meshCubeWith(posA, normA, texA, inxA, block, new Vector3f(wPosX, wPosY, wPosZ), new CubeSides(
                                        top(wPosX, wPosY, wPosZ).isTransparent(block),
                                        bottom(wPosX, wPosY, wPosZ).isTransparent(block),
                                        left(wPosX, wPosY, wPosZ).isTransparent(block),
                                        right(wPosX, wPosY, wPosZ).isTransparent(block),
                                        back(wPosX, wPosY, wPosZ).isTransparent(block),
                                        front(wPosX, wPosY, wPosZ).isTransparent(block)
                                ), light, new Vector3i(wPosX, wPosY, wPosZ));
                                break;
                            case CROSS:
                                CrossMesher.meshCrossWith(posA, normA, texA, inxA, block, new Vector3f(wPosX, wPosY, wPosZ), light, new Vector3i(wPosX, wPosY, wPosZ));
                        }
                    }
                }
            }
        }
    }
    public static void meshChunkWithBlend(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, ArrayList<Float> light, Vector3i loc) {
        for(int i = 0; i < Chunk.size; i++) {
            for(int j = 0; j < Chunk.size; j++) {
                for(int k = 0; k < Chunk.size; k++) {
                    int wPosX = i + loc.x*Chunk.size, wPosY = j + loc.y*Chunk.size, wPosZ = k + loc.z*Chunk.size;
                    Block block3;
                    if(wPosY-1 < 0) {
                        block3 = Blocks.AIR;
                    } else {
                        block3 = World.getBlockAtLazy(wPosX, wPosY-1, wPosZ);
                    }
                    Block block = World.getBlockAtLazy(wPosX, wPosY, wPosZ);
                    Block block2 = World.getBlockAtLazy(wPosX, wPosY+1, wPosZ);
                    Block block21 = World.getBlockAtLazy(wPosX+1, wPosY, wPosZ);
                    Block block22 = World.getBlockAtLazy(wPosX, wPosY, wPosZ+1);
                    Block block23 = World.getBlockAtLazy(wPosX-1, wPosY, wPosZ-1);
                    Block block24 = World.getBlockAtLazy(wPosX, wPosY, wPosZ);

                    if(block.visible() && block.isBlend()) {
                        switch(block.modelType()) {
                            case WATER:
                                WaterMesher.meshWaterWith(block2 == Blocks.AIR, block3 == Blocks.WATER, posA, normA, texA, inxA, block, new Vector3f(wPosX, wPosY, wPosZ), new CubeSides(
                                        top(wPosX, wPosY, wPosZ).isTransparent(block),
                                        bottom(wPosX, wPosY, wPosZ).isTransparent(block),
                                        left(wPosX, wPosY, wPosZ).isTransparent(block),
                                        right(wPosX, wPosY, wPosZ).isTransparent(block),
                                        back(wPosX, wPosY, wPosZ).isTransparent(block),
                                        front(wPosX, wPosY, wPosZ).isTransparent(block)
                                ), light, new Vector3i(wPosX, wPosY, wPosZ));
                                break;
                            case CUBE:
                                CubeMesher.meshCubeWith(posA, normA, texA, inxA, block, new Vector3f(wPosX, wPosY, wPosZ), new CubeSides(
                                        top(wPosX, wPosY, wPosZ).isTransparent(block),
                                        bottom(wPosX, wPosY, wPosZ).isTransparent(block),
                                        left(wPosX, wPosY, wPosZ).isTransparent(block),
                                        right(wPosX, wPosY, wPosZ).isTransparent(block),
                                        back(wPosX, wPosY, wPosZ).isTransparent(block),
                                        front(wPosX, wPosY, wPosZ).isTransparent(block)
                                ), light, new Vector3i(wPosX, wPosY, wPosZ));
                                break;
                            case CROSS:
                                CrossMesher.meshCrossWith(posA, normA, texA, inxA, block, new Vector3f(wPosX, wPosY, wPosZ), light, new Vector3i(wPosX, wPosY, wPosZ));
                        }
                    }
                }
            }
        }
    }
}
