package ity.opencraft.mesh;

import ity.opencraft.world.BlockSide;
import ity.opencraft.world.Block;
import ity.opencraft.world.World;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.stream.Collectors;

public class CrossMesher {
    public static int atlasSize = 4;

    private static void meshFace(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Vector3f> facePosA, ArrayList<Vector2f> texId, Vector3f faceNorm, ArrayList<Integer> faceInxA, ArrayList<Float> lightA, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        int bInx = posA.size();
        ArrayList<Vector3f> pos = new ArrayList<>(facePosA.stream().map(p -> p.add(loc)).collect(Collectors.toList()));
        ArrayList<Integer> inds = new ArrayList<>(faceInxA.stream().map(i -> i+bInx).collect(Collectors.toList()));
        ArrayList<Vector2f> texs = new ArrayList<>(texId.stream().map(v -> v.add(texOff).div(atlasSize,atlasSize)).collect(Collectors.toList()));
        posA.addAll(pos);
        inxA.addAll(inds);
        texA.addAll(texs);
        for(Vector3f p : pos) {
            normA.add(faceNorm);
        }
    }

    /**
     * Fill the given parameters with the mesh of a partial cube, translated.
     * @param posA The ArrayList to fill with positions
     * @param normA The ArrayList tof ill with normals
     * @param inxA The ArrayList to fill with indices
     * @param loc The offset by which to offset the generated mesh
     */
    public static void meshCrossWith(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Block block, Vector3f loc, ArrayList<Float> lightA, Vector3i wPos) {
        meshFace1(posA, normA, texA, inxA, loc, new Vector2f(block.textureIndex(BlockSide.CROSS_1, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))%atlasSize, block.textureIndex(BlockSide.LEFT, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))/atlasSize), lightA, wPos);
        meshFace2(posA, normA, texA, inxA, loc, new Vector2f(block.textureIndex(BlockSide.CROSS_2, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))%atlasSize, block.textureIndex(BlockSide.RIGHT, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))/atlasSize), lightA, wPos);
        meshFace3(posA, normA, texA, inxA, loc, new Vector2f(block.textureIndex(BlockSide.CROSS_2, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))%atlasSize, block.textureIndex(BlockSide.FRONT, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))/atlasSize), lightA, wPos);
        meshFace4(posA, normA, texA, inxA, loc, new Vector2f(block.textureIndex(BlockSide.CROSS_2, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))%atlasSize, block.textureIndex(BlockSide.BACK, new Vector3i((int)loc.x, (int)loc.y, (int)loc.z))/atlasSize), lightA, wPos);
    }

    private static void meshFace1(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Float> lightA, Vector3i wPos) {
        meshFace(posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f,  0.5f),
                new Vector3f(-0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f,  0.5f, -0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,1), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)), lightA, wPos);
    }
    private static void meshFace2(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Float> lightA, Vector3i wPos) {
        meshFace(posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f,  0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,-1), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)), lightA, wPos);
    }
    private static void meshFace3(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Float> lightA, Vector3i wPos) {
        meshFace(posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f,  0.5f),
                new Vector3f(-0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f,  0.5f, -0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,1), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)), lightA, wPos);
    }
    private static void meshFace4(ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Float> lightA, Vector3i wPos) {
        meshFace(posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f,  0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,-1), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)), lightA, wPos);
    }
}
