package ity.opencraft.mesh;

import ity.opencraft.world.Block;
import ity.opencraft.world.BlockSide;
import ity.opencraft.world.World;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;
import java.util.stream.Collectors;

public class WaterMesher {
    public static int atlasSize = 4;
    private static ArrayList<Vector2f> rotateLeftOnce(ArrayList<Vector2f> inp) {
        ArrayList<Vector2f> out = new ArrayList<>();
        for(int i = 1; i < inp.size(); i++) {
            out.add(inp.get(i));
        }
        out.add(inp.get(0));
        return out;
    }
    private static ArrayList<Vector2f> rotateLeft(ArrayList<Vector2f> inp, int n) {
        for(int i = 0; i < n; i++) {
            inp = rotateLeftOnce(inp);
        }
        return inp;
    }

    private static void meshFace(boolean randRotate, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, ArrayList<Vector3f> facePosA, ArrayList<Vector2f> texId, Vector3f faceNorm, ArrayList<Integer> faceInxA) {
        Random rand = new Random(loc.hashCode());
        int bInx = posA.size();
        ArrayList<Vector3f> pos = new ArrayList<>(facePosA.stream().map(p -> p.add(loc)).collect(Collectors.toList()));
        ArrayList<Integer> inds = new ArrayList<>(faceInxA.stream().map(i -> i+bInx).collect(Collectors.toList()));
        ArrayList<Vector2f> texs = new ArrayList<>(texId.stream().map(v -> v.add(texOff).div(atlasSize,atlasSize)).collect(Collectors.toList()));
        posA.addAll(pos);
        inxA.addAll(inds);
        if(randRotate) {
            texA.addAll(rotateLeft(texs, rand.nextInt(4)));
        } else {
            texA.addAll(texs);
        }
        for(Vector3f p : pos) {
            normA.add(faceNorm);
        }
    }

    private static void meshFaceFront(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y, wPos.z+1);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.45f, 0.5f),
                new Vector3f(-0.5f, -0.5f, 0.5f),
                new Vector3f( 0.5f, -0.5f, 0.5f),
                new Vector3f( 0.5f,  0.45f, 0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,1), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceBack(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y, wPos.z-1);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.45f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f,  0.45f, -0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,-1), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }
    private static void meshFaceLeft(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x-1, wPos.y, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.45f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f,  0.5f),
                new Vector3f(-0.5f,  0.45f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(-1,0,0), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceRight(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x+1, wPos.y, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f,  0.45f, -0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f,  0.45f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(1,0,0), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }
    private static void meshFaceTop2(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y+1, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f,  0.5f,  0.5f),
                new Vector3f( 0.5f,  0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(1f,0f),
                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f))), new Vector3f(0,1,0), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceTop(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y+1, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f,  0.45f, -0.5f),
                new Vector3f(-0.5f,  0.45f, -0.5f),
                new Vector3f(-0.5f,  0.45f,  0.5f),
                new Vector3f( 0.5f,  0.45f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(1f,0f),
                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f))), new Vector3f(0,1,0), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceBottom(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y-1, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f, -0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(1f,0f),
                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f))), new Vector3f(0,-1,0), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }
    private static void meshFaceBottom3(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y-1, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f, -0.5f-0.0625f, -0.5f),
                new Vector3f(-0.5f, -0.5f-0.0625f, -0.5f),
                new Vector3f(-0.5f, -0.5f-0.0625f,  0.5f),
                new Vector3f( 0.5f, -0.5f-0.0625f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(1f,0f),
                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f))), new Vector3f(0,-1,0), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }

    private static void meshFaceFront2(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y, wPos.z+1);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f, 0.5f),
                new Vector3f(-0.5f, -0.5f, 0.5f),
                new Vector3f( 0.5f, -0.5f, 0.5f),
                new Vector3f( 0.5f,  0.5f, 0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,1), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceBack2(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x, wPos.y, wPos.z-1);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f,  0.5f, -0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(0,0,-1), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }
    private static void meshFaceLeft2(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x-1, wPos.y, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f(-0.5f,  0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f, -0.5f),
                new Vector3f(-0.5f, -0.5f,  0.5f),
                new Vector3f(-0.5f,  0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(-1,0,0), new ArrayList<>(Arrays.asList(0, 1, 2,  2, 3, 0)));
    }
    private static void meshFaceRight2(boolean randFace, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Float> lightA, ArrayList<Integer> inxA, Vector3f loc, Vector2f texOff, Vector3i wPos) {
        float light = World.getLightAt(wPos.x+1, wPos.y, wPos.z);
        lightA.addAll(Arrays.asList(light, light, light, light));
        meshFace(randFace, posA, normA, texA, inxA, loc, texOff, new ArrayList<>(Arrays.asList(
                new Vector3f( 0.5f,  0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f, -0.5f),
                new Vector3f( 0.5f, -0.5f,  0.5f),
                new Vector3f( 0.5f,  0.5f,  0.5f))), new ArrayList<>(Arrays.asList(

                new Vector2f(0f,0f),
                new Vector2f(0f,1f),
                new Vector2f(1f,1f),
                new Vector2f(1f,0f))), new Vector3f(1,0,0), new ArrayList<>(Arrays.asList(2, 1, 0,  0, 3, 2)));
    }

    /**
     * Fill the given parameters with the mesh of a partial cube, translated.
     * @param posA The ArrayList to fill with positions
     * @param normA The ArrayList tof ill with normals
     * @param inxA The ArrayList to fill with indices
     * @param loc The offset by which to offset the generated mesh
     * @param sides The sides which to show. True means render, false means do not render.
     */
    public static void meshWaterWith(boolean isUp, boolean isBot, ArrayList<Vector3f> posA, ArrayList<Vector3f> normA, ArrayList<Vector2f> texA, ArrayList<Integer> inxA, Block block, Vector3f loc, CubeSides sides, ArrayList<Float> light, Vector3i wPos) {
        if(sides.back) {
            if(isUp) {
                meshFaceBack(block.textureRotate(BlockSide.BACK), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.BACK, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.BACK, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            } else {
                meshFaceBack2(block.textureRotate(BlockSide.BACK), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.BACK, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.BACK, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            }
        } if(sides.front) {
            if(isUp) {
                meshFaceFront(block.textureRotate(BlockSide.FRONT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.FRONT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.FRONT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            } else {
                meshFaceFront2(block.textureRotate(BlockSide.FRONT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.FRONT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.FRONT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            }
        } if(sides.top) {
            if(isUp) {
                meshFaceTop(block.textureRotate(BlockSide.TOP), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.TOP, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.TOP, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            } else {
                meshFaceTop2(block.textureRotate(BlockSide.TOP), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.TOP, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.TOP, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            }
        }
        if(sides.bottom) {
            meshFaceBottom(block.textureRotate(BlockSide.BOTTOM), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.BOTTOM, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.BOTTOM, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
        } if(sides.left) {
            if(isUp) {
                meshFaceLeft(block.textureRotate(BlockSide.LEFT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.LEFT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.LEFT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            } else {
                meshFaceLeft2(block.textureRotate(BlockSide.LEFT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.LEFT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.LEFT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);

            }
        } if(sides.right) {
            if(isUp) {
                meshFaceRight(block.textureRotate(BlockSide.RIGHT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.RIGHT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.RIGHT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            } else {
                meshFaceRight2(block.textureRotate(BlockSide.RIGHT), posA, normA, texA, light, inxA, loc, new Vector2f(block.textureIndex(BlockSide.RIGHT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) % atlasSize, block.textureIndex(BlockSide.RIGHT, new Vector3i((int) loc.x, (int) loc.y, (int) loc.z)) / atlasSize), wPos);
            }
        }
    }
}
