package ity.opencraft.render;

import ity.opencraft.data.*;
import ity.opencraft.io.Screen;
import ity.opencraft.io.Window;
import ity.opencraft.mesh.ChunkMesher;
import ity.opencraft.world.Blocks;
import ity.opencraft.world.Chunk;
import ity.opencraft.world.ChunkReference;
import ity.opencraft.backends.Backended;
import ity.opencraft.client.ViewportCamera;
import ity.opencraft.io.Windows;
import ity.opencraft.shader.Shader;
import ity.opencraft.util.MatrixUtils;
import ity.opencraft.util.TextureLoader;
import ity.opencraft.world.World;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;
import org.lwjgl.opengl.GL11;

import java.lang.reflect.Array;
import java.util.*;

// TODO clean up Renderer, it handles too much
public class Renderer extends Backended {
    private static IndexedMesh crosshair;
    private static HashMap<Vector3i, ChunkReference> chunks = new HashMap<>();
    private static ArrayList<Vector3i> toGen = new ArrayList<>();
    private static ArrayList<ChunkReference> toMesh = new ArrayList<>();
    private static ArrayList<ChunkReference> toUpload = new ArrayList<>();
    private static ArrayList<ChunkReference> toRender = new ArrayList<>();
    private static ArrayList<ChunkReference> toFree = new ArrayList<>();
    private static HashMap<Vector3i, ChunkReference> toDelete = new HashMap<>();
    private static ArrayList<Vector3i> toDeletePos = new ArrayList<>();
    private static Shader colorShader;
    private static Shader guiShader;
    private static Renderer instance;
    private static Texture atlas;
    private static final BlockMeshPool pool = new BlockMeshPool();

    public static ArrayList<Thread> worldGenWorkers = new ArrayList<>();

    private static HashMap<String, Long> perfTimes = new HashMap<>();
    private static long frameTime;

    // TODO: Renderer should not be backend-dependant just for clearing the screen.
    public void clearScreen() {
        ((Renderer)this.backend).clearScreen();
    }
    public void depthTest() {
        ((Renderer)this.backend).depthTest();
    }

    private static final int renderDistance = 4;
    private static final int renderDistanceY = 4;

    public static void remeshChunk(int x, int y, int z) {
        Vector3i vPos = new Vector3i(x, y, z);
        ChunkReference chunk = chunks.get(vPos);
        if(chunk == null || chunk.getMeshHandle() == null) return;
        chunk.getMeshHandle().free();
        ArrayList<Vector3f> pos = new ArrayList<>();
        ArrayList<Vector3f> norms = new ArrayList<>();
        ArrayList<Vector2f> texs = new ArrayList<>();
        ArrayList<Float> light = new ArrayList<>();
        ArrayList<Integer> inds = new ArrayList<>();
        ChunkMesher.meshChunkWithSolid(pos, norms, texs, inds, light, vPos);
        BlockMeshHandle mesh = pool.make(new BlockMeshData(pos, norms, texs, light, inds));
        chunk.setMeshHandle(mesh);
        chunk.setMeshData(mesh.data);
        ArrayList<Vector3f> pos2 = new ArrayList<>();
        ArrayList<Vector3f> norms2 = new ArrayList<>();
        ArrayList<Vector2f> texs2 = new ArrayList<>();
        ArrayList<Float> light2 = new ArrayList<>();
        ArrayList<Integer> inds2 = new ArrayList<>();
        ChunkMesher.meshChunkWithBlend(pos2, norms2, texs2, inds2, light2, vPos);
        BlockMeshHandle mesh2 = pool.make(new BlockMeshData(pos2, norms2, texs2, light2, inds2));
        chunk.setMeshHandle2(mesh2);
        chunk.setMeshData2(mesh2.data);
        toDeletePos.remove(vPos);
        toDelete.remove(vPos);
        for(ChunkReference chk : toRender) {
            if(chk.getPos().equals(vPos))
                return;
        }
        toRender.add(chunk);
    }

    // TODO clean up
    public static void doChunk() {
        ArrayList<Vector3i> toGenNew = new ArrayList<>();
        // TODO do in a circle around the player starting from the center, instead of in strips.
        for(int i = 0; i <= renderDistance*2; i++) {
            for(int j = renderDistanceY*2; j >= 0 ; j--) {
                for (int k = 0; k <= renderDistance*2; k++) {
                    Vector3i vPos = new Vector3i(((int)(ViewportCamera.pos.x+(0))/Chunk.size)+(i*((i%2)*2-1)/2),
                                                 ((int)(ViewportCamera.pos.y+(0))/Chunk.size)+(j*((j%2)*2-1)/2),
                                                 ((int)(ViewportCamera.pos.z+(0))/Chunk.size)+(k*((k%2)*2-1)/2));
                    if(vPos.x < 0 || vPos.y < 0|| vPos.z < 0) continue;
                    if(new Vector3f(vPos).sub(new Vector3f(ViewportCamera.pos).div(Chunk.size)).length() > renderDistance) continue;
                    toGenNew.add(vPos);
                }
            }
        }
        for(Vector3i vec : toGenNew) {
            if(!toGen.contains(vec)) {
                toDeletePos.remove(vec);
                toDelete.remove(vec);
                if(!chunks.containsKey(vec)) toGen.add(vec);
            }
        }
//        for(Vector3i vec : toGen) {
//            if(!toGenNew.contains(vec)) {
//                toDeletePos.add(vec);
//            }
//        }
        for(Map.Entry<Vector3i, Chunk> chunkEntry : World.chunks.entrySet()) {
            if(!toGenNew.contains(chunkEntry.getKey()) && !toDeletePos.contains(chunkEntry.getKey())) {
                Chunk chunk = chunkEntry.getValue();
                chunks.put(chunk.pos, World.getChunkAt(chunk.pos));
                toDeletePos.add(chunk.pos);
            }
        }
    }

    /**
     * Initialize the Renderer. Requires everything else to be ready.
     */
    public static void init() {
        instance = new Renderer();
        instance.depthTest();
//        for(int i = 0; i < 1; i++) {
//            Thread worker = new Thread(() -> {
//                while (true) {
//                    beginPerf("worldgen");
//                    worldGenUpdate();
//                    long taken = endPerf("worldgen") / 1000000;
//                    long max = 8;
//                    long toSleep = Math.max(max - taken, 0);
//                    try {
//                        Thread.sleep(toSleep);
//                    } catch (InterruptedException e) {
//                        throw new RuntimeException(e);
//                    }
//                }
//            }, "WorldGen worker "+i);
//            worldGenWorkers.add(worker);
//        }
        crosshair = new IndexedMesh();
        crosshair.init();
        Window win = Screen.get(Windows.MAIN.toString());
        float asp = (float)win.height/(float)win.width;
        crosshair.upload(new float[] {
                0.05f/asp, 0.05f, 0.0f,  // top right
                0.05f/asp, -0.05f, 0.0f,  // bottom right
                -0.05f/asp, -0.05f, 0.0f,  // bottom left
                -0.05f/asp, 0.05f, 0.0f   // top left
        }, new int[] {
                3, 1, 0,   // first triangle
                3, 2, 1    // second triangle
        });
        atlas = TextureLoader.load("atlas.png");
        Vector3i startPlayerPos = new Vector3i(World.sizeXZ/2,60/Chunk.size,World.sizeXZ/2);
        for(int i = startPlayerPos.x; i <= startPlayerPos.x+(renderDistance*2); i++) {
            for(int j = startPlayerPos.y; j <= startPlayerPos.y+(renderDistanceY*2); j++) {
                for (int k = startPlayerPos.z; k <= startPlayerPos.z+(renderDistance*2); k++) {
                    if(j < 0) continue;
                    Vector3i vPos = new Vector3i((i/2*((i%2)*2-1)), (j/2*((j%2)*2-1)), (k/2*((k%2)*2-1)));
                    if(new Vector3f(startPlayerPos).sub(new Vector3f(ViewportCamera.pos).div(Chunk.size)).length() > renderDistance) continue;
                    toGen.add(vPos);
                }
            }
        }

        colorShader = new Shader();
        colorShader.init("color.vs", "color.fs");
        guiShader = new Shader();
        guiShader.init("gui.vs", "gui.fs");
        ViewportCamera.pos.set(World.sizeXZ*Chunk.size/2+Chunk.size/2,80+128-70,World.sizeXZ*Chunk.size/2+Chunk.size/2);

        for(Thread worker : worldGenWorkers) {
            worker.start();
        }
    }

    private static synchronized void worldGenUpdate() {
        if(toGen.isEmpty()) {
            System.out.println("Possible deadlock");
            return;
        }
        Vector3i toGenPos = toGen.remove(0);
        if(toDelete.remove(toGenPos) != null) return;
        if(toDeletePos.remove(toGenPos)) return;
        ChunkReference chunk = World.getChunkAt(toGenPos.x, toGenPos.y, toGenPos.z);
        if(toRender.contains(chunk)) {
            // TODO fix
            return;
            //throw new IllegalStateException("Trying to generate a chunk that's being rendered");
        }
        toMesh.add(chunk);
        chunks.put(toGenPos, chunk);
    }

    private static void meshUpdate() {
        beginPerf("premesh");
        beginPerf("premesh.1");
        if(toMesh.isEmpty()) return;
        endPerf("premesh.1");
        beginPerf("premesh.2");
        ChunkReference toMeshChunk = toMesh.remove(0);
        if(!toMeshChunk.isChunkGenerated())
            System.out.println("Generating a chunk that should've been generated already.");
        endPerf("premesh.2");
        beginPerf("premesh.3");
        if(toRender.contains(toMeshChunk)) {
            throw new IllegalStateException("Trying to mesh a chunk that's being rendered");
        }
        endPerf("premesh.3");
        //if(toDelete.containsKey(toMeshChunk.pos)) return;
        beginPerf("premesh.4");
        BlockMeshData meshData = toMeshChunk.getMeshData();
        endPerf("premesh.4");
        beginPerf("premesh.5");
        if(meshData != null) {
            beginPerf("premesh.5.1");
            toUpload.add(toMeshChunk);
            endPerf("premesh.5.1");
            return;
        }
        endPerf("premesh.5");
        endPerf("premesh");
        beginPerf("solidmesh");
        ArrayList<Vector3f> pos = new ArrayList<>();
        ArrayList<Vector3f> norms = new ArrayList<>();
        ArrayList<Vector2f> texs = new ArrayList<>();
        ArrayList<Float> light = new ArrayList<>();
        ArrayList<Integer> inds = new ArrayList<>();
        ChunkMesher.meshChunkWithSolid(pos, norms, texs, inds, light, toMeshChunk.getPos());
        toMeshChunk.setMeshData(new BlockMeshData(pos, norms, texs, light, inds));
        endPerf("solidmesh");
        beginPerf("blendmesh");
        ArrayList<Vector3f> pos2 = new ArrayList<>();
        ArrayList<Vector3f> norms2 = new ArrayList<>();
        ArrayList<Vector2f> texs2 = new ArrayList<>();
        ArrayList<Float> light2 = new ArrayList<>();
        ArrayList<Integer> inds2 = new ArrayList<>();
        ChunkMesher.meshChunkWithBlend(pos2, norms2, texs2, inds2, light2, toMeshChunk.getPos());
        toMeshChunk.setMeshData2(new BlockMeshData(pos2, norms2, texs2, light2, inds2));
        toUpload.add(toMeshChunk);
        endPerf("blendmesh");
    }

    private static void uploadUpdate() {
        if(toUpload.isEmpty()) return;
        ChunkReference toUploadChunk = toUpload.remove(0);
        if(toRender.contains(toUploadChunk)) {
            throw new IllegalStateException("Trying to upload a chunk mesh of a chunk that's being rendered");
        }
        //if(toDelete.containsKey(toUploadChunk.pos)) return;
        if(toUploadChunk.getMeshHandle() != null) {
            toRender.add(toUploadChunk);
            return;
        }
        toUploadChunk.setMeshHandle(pool.make(toUploadChunk.getMeshData()));
        toUploadChunk.getMeshData().free();
        toUploadChunk.setMeshData(null);
        toUploadChunk.setMeshHandle2(pool.make(toUploadChunk.getMeshData2()));
        toUploadChunk.getMeshData2().free();
        toUploadChunk.setMeshData2(null);
        // toUploadChunk.meshData = null;
        // toUploadChunk.meshHandle.data = null;
        toRender.add(toUploadChunk);
    }

    private static void deleteUpdate() {
        for(Map.Entry<Vector3i, ChunkReference> pair : toDelete.entrySet()) {
            if(!toDeletePos.contains(pair.getKey()))
                toDeletePos.add(pair.getKey());
        }
        for(Vector3i vec : toDeletePos) {
            ChunkReference chunk = chunks.remove(vec);
            if(chunk != null) {
                toUpload.remove(chunk);
                if(toUpload.remove(chunk)) throw new IllegalStateException();
                toRender.remove(chunk);
                if(toRender.remove(chunk)) throw new IllegalStateException();
                toMesh.remove(chunk);
                if(toMesh.remove(chunk)) throw new IllegalStateException();
                toGen.remove(vec);
                if(toGen.remove(vec)) throw new IllegalStateException();
                toFree.add(chunk);
            }
        }
    }

    private static void freeUpdate() {
        if(toDelete.isEmpty())
            return;
        ChunkReference ref = toDelete.remove(0);
        Vector3i vec = ref.getPos();
        //World.freeChunkAt(vec.x, vec.y, vec.z);
    }

    private static void beginPerf(String name) {
        perfTimes.put(name, System.nanoTime());
    }

    private static long endPerf(String name) {
        long start = perfTimes.get(name);
        long end = System.nanoTime();
        long timeElapsed = end - start;
        System.out.print(name + ": ");
        System.out.println(timeElapsed);
        return timeElapsed;
    }

    // TODO: Do not hardcode Windows.MAIN
    /**
     * Render a frame
     * @param delta Delta Time, the time in miliseconds between the last and this frame.
     */
    public static void render(long delta) {
        beginPerf("frame");
        if(delta < 200) delta = 200;
        for(int j = 0; j <= 1; j++) {
            beginPerf("free");
            freeUpdate();
            endPerf("free");
            beginPerf("delete");
            deleteUpdate();
            endPerf("delete");
            beginPerf("worldgen");
            worldGenUpdate();
            endPerf("worldgen");
            beginPerf("mesh");
            meshUpdate();
            endPerf("mesh");
            beginPerf("upload");
            uploadUpdate();
            endPerf("upload");
            toDelete.clear();
        }
//        for(int i = 0; i <= (200-delta)/10; i++) {
//            new Thread(() -> {
//
//            }).run();
//        }
        beginPerf("framerender");
        beginPerf("prepass");
        instance.clearScreen();
        colorShader.bind();
        atlas.bind();
        colorShader.setUniform("iRes", new Vector2f(Screen.get(Windows.MAIN.toString()).width, Screen.get(Windows.MAIN.toString()).height));
        colorShader.setUniform("view", ViewportCamera.getViewportMatrix());
        colorShader.setUniform("proj", MatrixUtils.projectionMatrix(70f, ((float) Screen.get(Windows.MAIN.toString()).width) / ((float) Screen.get(Windows.MAIN.toString()).height), 0.1f, 1024));
        colorShader.setUniform("camPos", ViewportCamera.pos);
        if(World.getBlockAt((int) ViewportCamera.pos.x, (int) ViewportCamera.pos.y, (int) ViewportCamera.pos.z) == Blocks.WATER) {
            colorShader.setUniform("isWater", new Vector2f(1.0f, 2.5f));
            colorShader.setUniform("fogColor", new Vector3f(0.1f, 0.1f, 1.0f));
            GL11.glClearColor(0.1f, 0.1f, 1.0f, 1.0f);
        } else {
            colorShader.setUniform("isWater", new Vector2f(5.0f, 1.0f));
            colorShader.setUniform("fogColor", new Vector3f(0.5f, 0.85f, 1.0f));
            GL11.glClearColor(0.5f, 0.85f, 1.0f, 1.0f);
        }
        // TODO fix vertical fog
        colorShader.setUniform("renderDistance", (float)renderDistance);
        endPerf("prepass");
        beginPerf("render");
        for(int i = 0; i < 2; i++) {
            for (ChunkReference chunk : toRender) {
                if(i == 0) {
                    //beginPerf("pass1");
                    chunk.getMeshHandle().mesh.bind();
                    chunk.getMeshHandle().mesh.render();
                    chunk.getMeshHandle().mesh.unbind();
                    //endPerf("pass1");
                } else {
                    //beginPerf("pass2");
                    chunk.getMeshHandle2().mesh.bind();
                    chunk.getMeshHandle2().mesh.render();
                    chunk.getMeshHandle2().mesh.unbind();
                    //endPerf("pass2");
                }
            }
        }
        endPerf("render");
        beginPerf("ui");
        colorShader.unbind();
        atlas.unbind();
        guiShader.bind();
        guiShader.setUniform("iRes", new Vector2f(Screen.get(Windows.MAIN.toString()).width, Screen.get(Windows.MAIN.toString()).height));
        if(World.getBlockAt((int) ViewportCamera.pos.x, (int) ViewportCamera.pos.y, (int) ViewportCamera.pos.z) == Blocks.WATER) {
            guiShader.setUniform("blueAlpha", 0.6f);
            GL11.glDisable(GL11.GL_CULL_FACE);
        }
        else {
            guiShader.setUniform("blueAlpha", 0.0f);
            GL11.glEnable(GL11.GL_CULL_FACE);
        }
        Window win = Screen.get(Windows.MAIN.toString());
        float asp = (float)win.width/(float)win.height;
        crosshair.upload(new float[] {
                5f/asp, 5f, 0.0f,  // top right
                5f/asp, -5f, 0.0f,  // bottom right
                -5f/asp, -5f, 0.0f,  // bottom left
                -5f/asp, 5f, 0.0f   // top left
        }, new int[] {
                3, 1, 0,   // first triangle
                3, 2, 1    // second triangle
        });
        crosshair.bind();
        crosshair.render();
        crosshair.unbind();
        guiShader.unbind();
        endPerf("ui");
        //System.out.println(ViewportCamera.pos.x+", "+ViewportCamera.pos.y+", "+ViewportCamera.pos.z);
        endPerf("framerender");
        long thisFrameTime = endPerf("frame");
        if(frameTime == 0)
            frameTime = thisFrameTime;
        else {
            frameTime = (frameTime + thisFrameTime) / 2;
        }
        System.out.println("frametime: "+frameTime/1000000f);
        System.out.println("----------------------");
    }
}
