package ity.opencraft.shader;

import ity.opencraft.backends.Backended;
import org.joml.Matrix4f;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;

public class Shader extends Backended {
    /**
     * Initialize a shader
     * @param v The path to the vertex shader
     * @param f The path to the fragment shader
     * @return self
     */
    public Shader init(String v, String f) {
        return ((Shader)this.backend).init(v, f);
    }

    /**
     * Set a Matrix4f uniform
     * @param name Name of the uniform
     * @param matrix The value
     */
    public void setUniform(String name, Matrix4f matrix) {
        ((Shader)this.backend).setUniform(name, matrix);
    }

    /**
     * Set a Vector2f uniform
     * @param name Name of the uniform
     * @param vector The value
     */
    public void setUniform(String name, Vector2f vector) {
        ((Shader)this.backend).setUniform(name, vector);
    }

    /**
     * Set an int uniform
     * @param name Name of the uniform
     * @param value The value
     */
    public void setUniform(String name, int value) {
        ((Shader)this.backend).setUniform(name, value);
    }

    /**
     * Set a float uniform
     * @param name Name of the uniform
     * @param value The value
     */
    public void setUniform(String name, float value) {
        ((Shader)this.backend).setUniform(name, value);}

    /**
     * Set a Vector3i uniform
     * @param name Name of the uniform
     * @param vector The value
     */
    public void setUniform(String name, Vector3i vector) {
        ((Shader)this.backend).setUniform(name, vector);
    }

    /**
     * Set a Vector3f uniform
     * @param name Name of the uniform
     * @param value The value
     */
    public void setUniform(String name, Vector3f value) {
        ((Shader)this.backend).setUniform(name, value);
    }

    /**
     * Bind/Use the shader
     */
    public void bind() {
        ((Shader)this.backend).bind();
    }

    /**
     * Unbind/Unuse the shader
     */
    public void unbind() {
        ((Shader)this.backend).unbind();
    }
}
