package ity.opencraft.shader;

import ity.opencraft.backends.Backend;
import ity.opencraft.backends.BackendImpl;
import ity.opencraft.util.StringLoader;
import org.joml.Matrix4f;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.joml.Vector3i;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.system.MemoryUtil;

import java.nio.FloatBuffer;

import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL20.glGetUniformLocation;

@BackendImpl(Backend.GL)
public class ShaderGL extends Shader {

    private int pID;

    public Shader init(String v, String f) {
        pID = GL20.glCreateProgram();

        int VID = GL20.glCreateShader(GL20.GL_VERTEX_SHADER);

        GL20.glShaderSource(VID, StringLoader.loadResourceAsString("shaders/" + v));
        GL20.glCompileShader(VID);
        if (GL20.glGetShaderi(VID, GL20.GL_COMPILE_STATUS) == GL11.GL_FALSE) {
            System.err.println("Vertex Shader: " + GL20.glGetShaderInfoLog(VID));
        }

        int FID = GL20.glCreateShader(GL20.GL_FRAGMENT_SHADER);

        GL20.glShaderSource(FID, StringLoader.loadResourceAsString("shaders/" + f));
        GL20.glCompileShader(FID);
        if (GL20.glGetShaderi(FID, GL20.GL_COMPILE_STATUS) == GL11.GL_FALSE) {
            System.err.println("Fragment Shader: " + GL20.glGetShaderInfoLog(FID));
        }

        GL20.glAttachShader(pID, VID);
        GL20.glAttachShader(pID, FID);

        GL20.glLinkProgram(pID);
        if (GL20.glGetProgrami(pID, GL20.GL_LINK_STATUS) == GL11.GL_FALSE) {
            System.err.println("Program Linking: " + GL20.glGetProgramInfoLog(pID));
        }

        GL20.glValidateProgram(pID);
        if (GL20.glGetProgrami(pID, GL20.GL_VALIDATE_STATUS) == GL11.GL_FALSE) {
            System.err.println("Program Validation: " + GL20.glGetProgramInfoLog(pID));
        }

        return this;
    }

    public void setUniform(String name, Matrix4f matrix) {
        FloatBuffer matrixB = MemoryUtil.memAllocFloat(16);
        matrix.get(matrixB);
        glUniformMatrix4fv(glGetUniformLocation(pID, name), false, matrixB);
        MemoryUtil.memFree(matrixB);
    }

    public void setUniform(String name, Vector2f vector) {
        glUniform2fv(glGetUniformLocation(pID, name), new float[]{vector.x, vector.y});
    }

    public void setUniform(String name, int value) {
        glUniform1iv(glGetUniformLocation(pID, name), new int[]{value});
    }

    public void setUniform(String name, float value) {
        glUniform1fv(glGetUniformLocation(pID, name), new float[]{value});
    }

    public void setUniform(String name, Vector3i vector) {
        glUniform3iv(glGetUniformLocation(pID, name), new int[]{vector.x, vector.y, vector.z});
    }
    public void setUniform(String name, Vector3f vector) {
        glUniform3fv(glGetUniformLocation(pID, name), new float[]{vector.x, vector.y, vector.z});
    }

    public void bind() {
        glUseProgram(pID);
    }

    public void unbind() {
        glUseProgram(0);
    }
}