package ity.opencraft.util;

import ity.opencraft.backends.Backended;
import ity.opencraft.data.Texture;

import java.io.IOException;
import java.nio.ByteBuffer;

import static org.lwjgl.BufferUtils.createByteBuffer;

public class TextureLoader extends Backended {
    private static TextureLoader instance;

    /**
     * Initialize. Has to be called before load can be called.
     */
    public static void init() {
        instance = new TextureLoader();
    }

    /**
     * Load a texture
     * @param name The path to the texture, relative to the textures/ directory in resources.
     * @return The loaded texture
     */
    public static Texture load(String name) {
        return instance.loadTexture(name);
    }


    protected Texture loadTexture(String name) {
        return ((TextureLoader)backend).loadTexture(name);
    }

    /**
     * Load a resource into a buffer
     * @param resource The path to the resource
     * @param bufferSize The size of the buffer
     * @return The buffer
     */
    public static ByteBuffer ioResourceToByteBuffer(String resource, int bufferSize) throws IOException {
        ByteBuffer buffer = createByteBuffer(bufferSize);
        buffer.put(StringLoader.loadResourceAsStream(resource).readAllBytes());
        buffer.flip();
        return buffer;
    }
}