package ity.opencraft.world;

import ity.opencraft.mesh.BlockModelType;
import org.joml.Vector3i;

public abstract class Block {
    /**
     * Check if the block is transparent. Used to determine whether other block faces are visible through this block or not.
     * Examples include Air, Glass, Water...
     * @return Whether this block is transparent
     */
    public abstract boolean isTransparent(Block sideBlock);

    /**
     * Get the texture index of the block. Texture indices are counted from left to right and then from top to bottom.
     * @return The texture index of the block
     */
    public abstract int textureIndex(BlockSide side, Vector3i pos);

    /**
     * Check if the block is visible. An invisible block does not have any mesh or model, and is thus not rendered at all.
     * If you override this to return false, override isTransparent to return true.
     * @return Whether this block is transparent
     */
    public boolean visible() {
        return true;
    }

    public boolean textureRotate(BlockSide side) {
        return false;
    };

    public BlockModelType modelType() {
        return BlockModelType.CUBE;
    }

    public abstract boolean isSolid();

    public boolean isBlend() {
        return false;
    }
}
