package ity.opencraft.world;

import ity.opencraft.render.Renderer;
import ity.opencraft.util.OctavedNoise;
import ity.opencraft.world.blocks.BlockLeaves;
import org.joml.Vector3i;

import java.util.HashMap;
import java.util.Random;

// TODO negative coordinates
// As Y grows, it goes downwards instead of upwards.
public class World {
    /**
     * The size of the world in the X and Z direction; The horizontal size.
     */
    public static final int sizeXZ = 1024;
    public static final int sizeY = 64;

    private static HashMap<String, Long> perfTimes = new HashMap<>();

    public static HashMap<Vector3i, Chunk> chunks = new HashMap<>();
    private static HashMap<Vector3i, ChunkReference> chunkRefs = new HashMap<>();


    private static Random rand = new Random();

    private static final int mult = 128;
    private static final double multScale = 64.0;
    private static final int height = -128;

    private static int random(int x, int y) {
        Random pRand = new Random(((x * 435747689L) & 438) ^ y);

        return pRand.nextInt(10000);
    }

    private static void beginPerf(String name) {
        perfTimes.put(name, System.nanoTime());
    }

    private static void endPerf(String name) {
        long start = perfTimes.get(name);
        long end = System.nanoTime();
        long timeElapsed = end - start;
        System.out.print(name + ": ");
        System.out.println(timeElapsed);
    }

    private static boolean isSkyVisible(int x, int y, int z) {
        while(isChunkAt(x/Chunk.size,y/Chunk.size,z/Chunk.size)) {
            if(!World.getBlockAt(x, y, z).isTransparent(Blocks.AIR)) return false;
            y += 1;
        }
        return true;
    }

    private static Block makeBlockAt(int x, int y, int z) {

        double mult1 = (OctavedNoise.octavedNoise((x+63156) / 4096f, 0, (z+6533) / 4096f, 1, 32, 0.4f) + 1) * 0.5f;
        double mult = (OctavedNoise.octavedNoise((x+9276) / 256f, 0, (z+4317) / 256f, 1, 32, 0.4f) + 1) * 32f;

        int oY = y;
        y += height*mult1;

        //System.out.println(mult);

        double noise      = OctavedNoise.octavedNoise(x / multScale, y       / multScale, z / multScale, 3, 32, 0.4f) * mult;
        double noiseBelow = OctavedNoise.octavedNoise(x / multScale, (y - 1) / multScale, z / multScale, 3, 32, 0.4f) * mult;

        /*
        double treeNoise  = random(x, z);
        if(treeNoise > 9908) {
            if (!(noiseBelow  < (y - 1) - mult) && noise       < (y - 1) - (mult - 1)) return Blocks.WOOD_LOG;

            double noiseBelow2 = OctavedNoise.octavedNoise(x / multScale, (y - 2) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow2 < (y - 2) - mult) && noiseBelow  < (y - 2) - (mult - 2)) return Blocks.WOOD_LOG;

            double noiseBelow3 = OctavedNoise.octavedNoise(x / multScale, (y - 3) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow3 < (y - 3) - mult) && noiseBelow2 < (y - 3) - (mult - 3)) return Blocks.LEAVES;
        }

        double treeNoiseXP  = random(x+1, z);
        if(treeNoiseXP > 9908) {
            double noiseBelow2XP = OctavedNoise.octavedNoise((x + 1) / multScale, (y - 2) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            double noiseBelowXP = OctavedNoise.octavedNoise((x + 1) / multScale, (y - 1) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow2XP < (y - 2) - mult) && noiseBelowXP < (y - 2) - (mult - 2)) return Blocks.LEAVES;
        }

        double treeNoiseXN  = random(x-1, z);
        if(treeNoiseXN > 9908) {
            double noiseBelow2XN = OctavedNoise.octavedNoise((x - 1) / multScale, (y - 2) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            double noiseBelowXN = OctavedNoise.octavedNoise((x - 1) / multScale, (y - 1) / multScale, z / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow2XN < (y - 2) - mult) && noiseBelowXN < (y - 2) - (mult - 2)) return Blocks.LEAVES;
        }

        double treeNoiseZP  = random(x, z+1);
        if(treeNoiseZP > 9908) {
            double noiseBelow2ZP = OctavedNoise.octavedNoise(x / multScale, (y - 2) / multScale, (z + 1) / multScale, 3, 32, 0.4f) * mult;
            double noiseBelowZP = OctavedNoise.octavedNoise(x / multScale, (y - 1) / multScale, (z + 1) / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow2ZP < (y - 2) - mult) && noiseBelowZP < (y - 2) - (mult - 2)) return Blocks.LEAVES;
        }

        double treeNoiseZN  = random(x, z-1);
        if(treeNoiseZN > 9908) {
            double noiseBelow2ZN = OctavedNoise.octavedNoise(x / multScale, (y - 2) / multScale, (z - 1) / multScale, 3, 32, 0.4f) * mult;
            double noiseBelowZN = OctavedNoise.octavedNoise(x / multScale, (y - 1) / multScale, (z - 1) / multScale, 3, 32, 0.4f) * mult;
            if (!(noiseBelow2ZN < (y - 2) - mult) && noiseBelowZN  < (y - 2) - (mult - 2)) return Blocks.LEAVES;
        }
        */

        double noiseAbove  = OctavedNoise.octavedNoise(x / multScale, (y + 1) / multScale, z / multScale, 3, 32, 0.4f) * mult;

        int waterLevel = 104;

        if(!(noiseBelow<(y-1)-mult) && noise<(y-1)-(mult-1) && rand.nextInt(5) == 0) {
            if(oY <= waterLevel-1)
                return Blocks.WATER;
            else
                return Blocks.TALL_GRASS;
        }
        if(!(noiseBelow<(y-1)-mult) && noise<(y-1)-(mult-1) && rand.nextInt(20) == 0) {
            if(oY <= waterLevel-1)
                return Blocks.WATER;
            else
                return Blocks.DANDELION;
        }
        if(noise<y-mult) {
            if(oY < waterLevel)
                return Blocks.WATER;
            else
                return Blocks.AIR;
        } else if(noiseAbove<y-(mult-1)) {
            if(oY < waterLevel-1)
                return Blocks.DIRT;
            else
                return Blocks.GRASS;
        } else if(noiseAbove<y-(mult-2)) {
            return Blocks.DIRT;
        } else if(rand.nextInt(100) == 0) {
            return Blocks.COAL_ORE;
        } else {
            return Blocks.STONE;
        }
    }

    private static Chunk makeChunkAt(int x, int y, int z) {
        beginPerf("chunkgen");
        System.out.println("Generating chunk x:"+x+", y:"+y+", z:"+z);
        Chunk chunk = new Chunk(new Vector3i(x, y, z));
        for(int i = 0; i < Chunk.size; i++) {
            for(int j = 0; j < Chunk.size; j++) {
                for(int k = 0; k < Chunk.size; k++) {
                    chunk.setBlockAt(i,j,k, makeBlockAt(x*Chunk.size+i, y*Chunk.size+j, z*Chunk.size+k));
                    Vector3i pos = new Vector3i(x, y, z);
                    //chunk.setLightAt(i, j, k, isSkyVisible(x, y, z) ? 1.0f : 0.3f);
                }
            }
        }
        endPerf("chunkgen");
        return chunk;
    }

    private static void makeLightAt(Chunk chunk, int x, int y, int z) {
        for(int i = 0; i < Chunk.size; i++) {
            for(int j = 0; j < Chunk.size; j++) {
                for(int k = 0; k < Chunk.size; k++) {
                    chunk.setLightAt(i, j, k, isSkyVisible(x*Chunk.size+i, y*Chunk.size+j, z*Chunk.size+k) ? 1.0f : 0.1f);
                }
            }
        }
    }

    /**
     * Get a block in the world at a given position. Takes the world position of the block.
     * @param x The x position
     * @param y The y position
     * @param z The z position
     * @return The block at the given position
     */
    public static Block getBlockAt(int x, int y, int z) {
        Chunk chunk = getChunkAtBlockPosRaw(x,y,z);
        Block block = chunk.getBlockAt(x%Chunk.size, y%Chunk.size, z%Chunk.size);
        // TODO better handling
        //if(block == null) throw new RuntimeException("Block at "+x+","+y+","+z+" was NULL.");
        return block;
    }

    public static Block getBlockAtLazy(int x, int y, int z) {
        if(isChunkAtBlockPos(x, y, z)) {
            Chunk chunk = getChunkAtBlockPosRaw(x, y, z);
            if(chunk == null)
                System.out.println(isChunkAtBlockPos(x, y, z));
            Block block = chunk.getBlockAt(x % Chunk.size, y % Chunk.size, z % Chunk.size);
            return block;
        } else {
            Block block = makeBlockAt(x, y, z);
            return block;
        }
    }

    public static float getLightAt(int x, int y, int z) {
        //Chunk chunk = getChunkAtBlockPosRaw(x,y,z);
        //return 1.0f;
        //return chunk.getLightAt(x%Chunk.size, y%Chunk.size, z%Chunk.size);
        // TODO better handling
        //if(block == null) throw new RuntimeException("Block at "+x+","+y+","+z+" was NULL.");
        return 1.0f;
    }

    public static void setLightAt(int x, int y, int z, float level) {
        Chunk chunk = getChunkAtBlockPosRaw(x,y,z);
        chunk.setLightAt(x%Chunk.size, y%Chunk.size, z%Chunk.size, level);
        // TODO better handling
        //if(block == null) throw new RuntimeException("Block at "+x+","+y+","+z+" was NULL.");
    }

    /**
     * Set a block in the world at a given position. Takes the world position of the block.
     * @param x The x position
     * @param y The y position
     * @param z The z position
     * @param block The block to put
     */
    public static void setBlockAt(int x, int y, int z, Block block) {
        Chunk chunk = getChunkAtBlockPosRaw(x,y,z);
        Vector3i blockInChunk = new Vector3i(x%Chunk.size, y%Chunk.size, z%Chunk.size);
        Vector3i chunkinWorld = new Vector3i(x/Chunk.size, y/Chunk.size, z/Chunk.size);
        chunk.setBlockAt(blockInChunk.x, blockInChunk.y, blockInChunk.z, block);
        Renderer.remeshChunk(chunkinWorld.x, chunkinWorld.y, chunkinWorld.z);
        // TODO will fail on world border
        if(blockInChunk.x == 0)
            Renderer.remeshChunk(chunkinWorld.x-1, chunkinWorld.y, chunkinWorld.z);
        if(blockInChunk.x == Chunk.size-1)
            Renderer.remeshChunk(chunkinWorld.x+1, chunkinWorld.y, chunkinWorld.z);
        if(blockInChunk.y == 0)
            Renderer.remeshChunk(chunkinWorld.x, chunkinWorld.y-1, chunkinWorld.z);
        if(blockInChunk.y == Chunk.size-1)
            Renderer.remeshChunk(chunkinWorld.x, chunkinWorld.y+1, chunkinWorld.z);
        if(blockInChunk.z == 0)
            Renderer.remeshChunk(chunkinWorld.x, chunkinWorld.y, chunkinWorld.z-1);
        if(blockInChunk.z == Chunk.size-1)
            Renderer.remeshChunk(chunkinWorld.x, chunkinWorld.y, chunkinWorld.z+1);
    }

    /**
     * Get a chunk at a given chunk position. Takes a location in *chunks*.
     * @param x The x chunk position
     * @param y The y chunk position
     * @param z The z chunk position
     * @return The chunk at the given position
     */
    public static ChunkReference getChunkAt(int x, int y, int z) {
        Vector3i key = new Vector3i(x,y,z);
        ChunkReference ref = chunkRefs.get(key);
        if(ref == null) {
            ref = new ChunkReference(x, y, z);
            chunkRefs.put(key, ref);
        }
        if(!isChunkAt(key)) {
            Chunk chunk = makeChunkAt(x, y, z);
            if(chunk == null)
                throw new IllegalStateException("Failed to create chunk");
            setChunkAt(x, y, z, chunk);
        }
        return ref;
    }

    public static ChunkReference getChunkAt(Vector3i pos) {
        return getChunkAt(pos.x, pos.y, pos.z);
    }
    public static boolean isChunkAt(int x, int y, int z) {
        return !(chunks.get(new Vector3i(x,y,z)) == null);
    }
    public static boolean isChunkAt(Vector3i pos) {
        return isChunkAt(pos.x, pos.y, pos.z);
    }
    private static Chunk getChunkAtRaw(int x, int y, int z) {
        if(isChunkAt(x, y, z)) {
            return chunks.get(new Vector3i(x,y,z));
        } else {
            Chunk chunk = makeChunkAt(x, y, z);
            setChunkAt(x, y, z, chunk);
            //makeLightAt(chunk, x, y, z);
            if(chunk == null)
                throw new IllegalStateException("Failed to generate chunk");
            return chunk;
        }
    }

    // TODO refactor, package private is ugly.
    static Chunk getChunkAtRaw(Vector3i pos) {
        return getChunkAtRaw(pos.x, pos.y, pos.z);
    }

    /**
     * Get a chunk at a given block position. Takes a block position and returns the chunk the given block position is in.
     * @param x The x block position
     * @param y The y block position
     * @param z The z block position
     * @return The chunk containing the given block position
     */
    public static ChunkReference getChunkAtBlockPos(int x, int y, int z) {
        return getChunkAt(x/Chunk.size,y/Chunk.size,z/Chunk.size);
    }

    private static Chunk getChunkAtBlockPosRaw(int x, int y, int z) {
        return getChunkAtRaw(x/Chunk.size,y/Chunk.size,z/Chunk.size);
    }

    private static boolean isChunkAtBlockPos(int x, int y, int z) {
        return isChunkAt(x/Chunk.size,y/Chunk.size,z/Chunk.size);
    }

    /**
     * Replace a chunk at a given chunk location.
     * @param x The x chunk position
     * @param y The y chunk position
     * @param z The z chunk position
     * @param chunk The chunk to replace with
     */
    private static void setChunkAt(int x, int y, int z, Chunk chunk) {
        if(chunk == null) {
            throw new IllegalStateException("Trying to insert null as a chunk");
        }
        chunks.put(new Vector3i(x,y,z), chunk);
    }

    public static void freeChunkAt(int x, int y, int z) {
        Chunk chunk = chunks.remove(new Vector3i(x,y,z));
        if(chunk != null) {
            chunk.free();
            chunkRefs.remove(new Vector3i(x, y, z)).isChunkPresent = false;
        }
    }

    public static int allocatedChunks() {
        return chunks.size();
    }
}
