#version 460

in vec3 passNorm;
in vec2 passTexUv;
in vec3 fragPos;
in float passLight;

out vec4 color;

uniform sampler2D tex;
uniform vec2 iRes;
uniform vec3 camPos;
uniform float renderDistance;
uniform vec2 isWater;
uniform vec3 fogColor;

const float eps = 0.001;

bool eqFlt(float a, float b) {
    return abs(a - b) < eps;
}

bool eqVec(vec3 a, vec3 b) {
    return eqFlt(a.x,b.x) && eqFlt(a.y,b.y) && eqFlt(a.z,b.z);
}

const float fogThick = 1.0;
const float chunkSize = 32;

void main() {
    vec4 smp = texture(tex, passTexUv);
    vec3 col = texture(tex, passTexUv).rgb * ((passLight/2.0)+0.5);
    if(eqVec(passNorm, vec3(0,-1,0)))
        col *= 0.5;
    else if(eqVec(passNorm, vec3(1,0,0)) || eqVec(passNorm, vec3(-1,0,0)))
        col *= 0.7;
    else if(eqVec(passNorm, vec3(0,0,1)) || eqVec(passNorm, vec3(0,0,-1)))
        col *= 0.9;
    vec3 toObj = fragPos - camPos;
    float dist = length(toObj)/((renderDistance/isWater.y-fogThick)*chunkSize);
    //col = mix(col, vec3(0.5f, 0.85f, 1.0f), min(1.0, pow(dist, 5)));
    col = mix(col, fogColor, min(1.0, pow(dist, isWater.x)));
    color = vec4(col, smp.a);
    if(smp.a < 0.1f) discard;
}
